@@ -3,7 +3,7 @@
 
 /*
  *  Main header file for the ALSA driver
- *  Copyright (c) 1994-2000 by Jaroslav Kysela <perex@suse.cz>
+ *  Copyright (c) 1994-2000 by Jaroslav Kysela <perex@perex.cz>
  *
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -22,6 +22,16 @@
  *
  */
 
+#include "config.h"
+#include <linux/module.h>
+
+/* number of supported soundcards */
+#ifdef CONFIG_SND_DYNAMIC_MINORS
+#define SNDRV_CARDS 32
+#else
+#define SNDRV_CARDS 8		/* don't change - minor numbers */
+#endif
+
 #include <linux/version.h>
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
 #include <linux/autoconf.h>
@@ -264,6 +274,10 @@ static inline struct proc_dir_entry *PDE(const struct inode *inode)
 typedef unsigned __nocast gfp_t;
 #endif
 
+#ifndef CONFIG_HAVE_GFP_DMA32
+#define GFP_DMA32 0		/* driver must check for 32-bit address */
+#endif
+
 #include <linux/wait.h>
 #ifndef wait_event_timeout
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
@@ -941,7 +955,7 @@ int snd_pci_dev_present(const struct pci_device_id *ids);
 /*
  * memory allocator wrappers
  */
-#ifdef CONFIG_SND_DEBUG_MEMORY
+#if defined(CONFIG_SND_DEBUG_MEMORY) && !defined(SKIP_HIDDEN_MALLOCS)
 
 #include <linux/slab.h>
 void *snd_hidden_kmalloc(size_t size, gfp_t flags);
@@ -1165,15 +1179,15 @@ static inline int snd_pci_compat_register_driver(struct pci_driver *driver)
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
 #ifdef CONFIG_PCI
 #ifndef CONFIG_HAVE_NEW_PCI_SAVE_STATE
-void snd_pci_compat_save_state(struct pci_dev *pci);
-void snd_pci_compat_restore_state(struct pci_dev *pci);
-static inline void snd_pci_orig_save_state(struct pci_dev *pci, u32 *buffer)
+int snd_pci_compat_save_state(struct pci_dev *pci);
+int snd_pci_compat_restore_state(struct pci_dev *pci);
+static inline int snd_pci_orig_save_state(struct pci_dev *pci, u32 *buffer)
 {
-	pci_save_state(pci, buffer);
+	return pci_save_state(pci, buffer);
 }
-static inline void snd_pci_orig_restore_state(struct pci_dev *pci, u32 *buffer)
+static inline int snd_pci_orig_restore_state(struct pci_dev *pci, u32 *buffer)
 {
-	pci_restore_state(pci, buffer);
+	return pci_restore_state(pci, buffer);
 }
 
 #undef pci_save_state
@@ -1362,7 +1376,7 @@ static inline int snd_pci_enable_msi(struct pci_dev *dev) { return -1; }
 #endif
 
 /* kmemdup() wrapper */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19) || defined(CONFIG_SND_DEBUG_MEMORY)
 #include <linux/string.h>
 static inline void *snd_kmemdup(const void *src, size_t len, gfp_t gfp)
 {
@@ -1404,7 +1418,7 @@ extern struct pt_regs *snd_irq_regs;
 #define CONFIG_SYSFS_DEPRECATED	1
 #endif
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 21)
+#ifndef CONFIG_HAVE_IS_POWER_OF_2
 static inline __attribute__((const))
 int is_power_of_2(unsigned long n)
 {
@@ -1412,4 +1426,130 @@ int is_power_of_2(unsigned long n)
 }
 #endif
 
+#ifdef CONFIG_PCI
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 23)
+#define snd_pci_revision(pci)	((pci)->revision)
+#else
+#include <linux/pci.h>
+static inline unsigned char snd_pci_revision(struct pci_dev *pci)
+{
+	unsigned char rev;
+	pci_read_config_byte(pci, PCI_REVISION_ID, &rev);
+	return rev;
+}
+#endif
+#endif /* PCI */
+
+/* BIT* macros */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
+#include <linux/bitops.h>
+/*
+#ifndef BIT
+#define BIT(nr)			(1UL << (nr))
+#endif
+*/
+#ifndef BIT_MASK
+#define BIT_MASK(nr)		(1UL << ((nr) % BITS_PER_LONG))
+#endif
+#ifndef BIT_WORD
+#define BIT_WORD(nr)		((nr) / BITS_PER_LONG)
+#endif
+#ifndef BITS_PER_BYTE
+#define BITS_PER_BYTE		8
+#endif
+#endif
+
+#ifndef CONFIG_HAVE_FFS
+#if defined(__i386__)
+static inline unsigned long __ffs(unsigned long word)
+{
+	__asm__("bsfl %1,%0"
+		:"=r" (word)
+		:"rm" (word));
+	return word;
+}
+#else
+static inline unsigned long __ffs(unsigned long word)
+{
+	__asm__("need_asm_for_your_arch_in_adriver.h");
+	return word;
+}
+#endif
+#endif
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 22)
+#ifndef uninitialized_var
+#define uninitialized_var(x) x = x
+#endif
+#endif /* <2.6.0 */
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
+typedef unsigned long uintptr_t;
+#endif
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 16)
+#include <linux/time.h>
+#define do_posix_clock_monotonic_gettime getnstimeofday
+#endif
+
+/* undefine SNDRV_CARDS again - it'll be re-defined in sound/core.h */
+#undef SNDRV_CARDS
+
+/* put_unaligned_*() */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
+#include <asm/unaligned.h>
+static inline void put_unaligned_le16(u16 val, void *p)
+{
+	val = cpu_to_le16(val);
+	put_unaligned(val, (u16 *)p);
+}
+
+static inline void put_unaligned_le32(u32 val, void *p)
+{
+	val = cpu_to_le32(val);
+	put_unaligned(val, (u32 *)p);
+}
+
+static inline void put_unaligned_le64(u64 val, void *p)
+{
+	val = cpu_to_le64(val);
+	put_unaligned(val, (u64 *)p);
+}
+
+static inline void put_unaligned_be16(u16 val, void *p)
+{
+	val = cpu_to_be16(val);
+	put_unaligned(val, (u16 *)p);
+}
+
+static inline void put_unaligned_be32(u32 val, void *p)
+{
+	val = cpu_to_be32(val);
+	put_unaligned(val, (u32 *)p);
+}
+
+static inline void put_unaligned_be64(u64 val, void *p)
+{
+	val = cpu_to_be64(val);
+	put_unaligned(val, (u64 *)p);
+}
+#endif
+
+/* list_first_entry */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 22)
+#include <linux/list.h>
+#ifndef list_first_entry
+#define list_first_entry(ptr, type, member) \
+	list_entry((ptr)->next, type, member)
+#endif
+#endif /* < 2.6.22 */
+
+#ifndef upper_32_bits
+#define upper_32_bits(n) ((u32)(((n) >> 16) >> 16))
+#endif
+
+#ifndef CONFIG_HAVE_PAGE_TO_PFN
+#define page_to_pfn(page)       (page_to_phys(page) >> PAGE_SHIFT)
+#endif
+
 #endif /* __SOUND_LOCAL_DRIVER_H */
