@@ -25,7 +25,9 @@ Extended Message Function With Response Cacheing
 #include "hpimsgx.h"
 #include "hpidebug.h"
 
-extern struct pci_device_id asihpi_pci_tbl[];	/* hpimod.c mod device tbl */
+static struct pci_device_id asihpi_pci_tbl[] = {
+#include "hpipcida.h"
+};
 
 static struct hpios_spinlock msgxLock;
 
@@ -73,7 +75,7 @@ static inline void HW_EntryPoint(
 	if (phm->wAdapterIndex < HPI_MAX_ADAPTERS) {
 		ep = (HPI_HandlerFunc *) hpi_entry_points[phm->wAdapterIndex];
 		if (ep) {
-			HPI_DEBUG_MESSAGE(phm);
+			HPI_DEBUG_MESSAGE(DEBUG, phm);
 			ep(phm, phr);
 			HPI_DEBUG_RESPONSE(phr);
 			return;
@@ -419,7 +421,7 @@ void HPI_MessageEx(
 	void *hOwner
 )
 {
-	HPI_DEBUG_MESSAGE(phm);
+	HPI_DEBUG_MESSAGE(DEBUG, phm);
 
 	if (phm->wType != HPI_TYPE_MESSAGE) {
 		HPI_InitResponse(phr, phm->wObject, phm->wFunction,
@@ -465,8 +467,7 @@ void HPI_MessageEx(
 		void *ep = NULL;
 		char *ep_name;
 
-		hpi_debug_message(phm, HPI_DEBUG_FLAG_DEBUG
-			FILE_LINE DBG_TEXT("DEBUG "));
+		HPI_DEBUG_MESSAGE(ERROR, phm);
 
 		ep = hpi_entry_points[phm->wAdapterIndex];
 
@@ -476,13 +477,11 @@ void HPI_MessageEx(
 			ep_name = "HPI_6000";
 		else if (ep == HPI_6205)
 			ep_name = "HPI_6205";
-		else if (ep == HPI_4000)
-			ep_name = "HPI_4000";
 		else
 			ep_name = "unknown";
 
 		HPI_DEBUG_LOG(ERROR,
-			DBG_TEXT("HPI %s Response - error# %d\n"),
+			"HPI %s Response - error# %d\n",
 			ep_name, phr->wError);
 
 		if (hpiDebugLevel >= HPI_DEBUG_LEVEL_VERBOSE)
@@ -598,7 +597,8 @@ static void InStreamClose(
 	if (hOwner ==
 		aIStreamUserOpen[phm->wAdapterIndex][phm->u.d.wStreamIndex].
 		hOwner) {
-		/* HPI_DEBUG_LOG(INFO,"closing adapter %d instream %d owned by %p\n",
+		/* HPI_DEBUG_LOG(INFO,"closing adapter %d "
+		   "instream %d owned by %p\n",
 		   phm->wAdapterIndex, phm->u.d.wStreamIndex, hOwner); */
 		aIStreamUserOpen[phm->wAdapterIndex][phm->u.d.wStreamIndex].
 			hOwner = NULL;
@@ -701,7 +701,8 @@ static void OutStreamClose(
 	if (hOwner ==
 		aOStreamUserOpen[phm->wAdapterIndex][phm->u.d.wStreamIndex].
 		hOwner) {
-		/* HPI_DEBUG_LOG(INFO,"closing adapter %d outstream %d owned by %p\n",
+		/* HPI_DEBUG_LOG(INFO,"closing adapter %d "
+		   "outstream %d owned by %p\n",
 		   phm->wAdapterIndex, phm->u.d.wStreamIndex, hOwner); */
 		aOStreamUserOpen[phm->wAdapterIndex][phm->u.d.wStreamIndex].
 			hOwner = NULL;
@@ -898,14 +899,15 @@ static u16 HPIMSGX_Init(
 		HPI_LookupEntryPointFunction(phm->u.s.Resource.r.Pci);
 
 	if (entry_point_func) {
-		HPI_DEBUG_MESSAGE(phm);
+		HPI_DEBUG_MESSAGE(DEBUG, phm);
 		entry_point_func(phm, &hr);
 	} else {
 		phr->wError = HPI_ERROR_PROCESSING_MESSAGE;
 		return phr->wError;
 	}
-	/* if the adapter was created succesfully save the mapping for future use */
 	if (hr.wError == 0) {
+		/* the adapter was created succesfully
+		   save the mapping for future use */
 		hpi_entry_points[hr.u.s.wAdapterIndex] = entry_point_func;
 		/* prepare adapter (pre-open streams etc.) */
 		HPI_DEBUG_LOG(DEBUG,
@@ -936,7 +938,7 @@ static void HPIMSGX_Cleanup(
 	}
 
 	for (; wAdapter < wAdapterLimit; wAdapter++) {
-		/*	printk(KERN_INFO "Cleanup adapter #%d\n",wAdapter); */
+		/*      printk(KERN_INFO "Cleanup adapter #%d\n",wAdapter); */
 		for (i = 0; i < HPI_MAX_STREAMS; i++) {
 			if (hOwner == aOStreamUserOpen[wAdapter][i].hOwner) {
 				struct hpi_message hm;
